/*
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * Trap handlers for m68k processors
 *
 * Contributors:
 *      (originally from Sun 3 code)
 *      Bryan Ford
 *      Niklas Hallqvist
 *      Markus Wild
 */

#include <mach_kdb.h>

#include <mach/exception.h>
#include <mach/kern_return.h>
#include <mach/vm_param.h>

#include <kern/thread.h>
#include <kern/task.h>
#include <kern/timer.h>
#include <kern/syscall_sw.h>
#include <kern/time_stamp.h>
#include <kern/ast.h>

#include <vm/vm_map.h>
#include <vm/vm_fault.h>
#include <vm/vm_kern.h>

#include <machine/frame.h>
#include <machine/mach_param.h>
#include <machine/trap.h>


static short sfsizes[16] = {
  8 + FMT0SIZE, 8 + FMT1SIZE, 8 + FMT2SIZE, 0, 0, 0, 0, 0, 0, 8 + FMT9SIZE,
  8 + FMTASIZE, 8 + FMTBSIZE, 0, 0, 0, 0,
};


char *trap_type[] = {
  "Reset SP",
  "Reset PC",
  "Bus error",
  "Address error",
  "Illegal instruction",
  "Divide by zero",
  "CHK, CHK2 instruction",
  "TRAPV, cpTRAPcc, cpTRAPcc instruction",
  "Privilege violation",
  "Trace",
  "1010 emulator trap",
  "1111 emulator trap",
  "Vector address 0x30",
  "Coprocessor protocol error",
  "Stack format error",
  "Uninitialized interrupt",
  "Vector address 0x40",
  "Vector address 0x44",
  "Vector address 0x48",
  "Vector address 0x4c",
  "Vector address 0x50",
  "Vector address 0x54",
  "Vector address 0x58",
  "Vector address 0x5c",
  "Spurious interrupt",
};
int     TRAP_TYPES = sizeof (trap_type) / sizeof (trap_type[0]);

#if     MACH_KDB
boolean_t debug_all_traps_with_kdb = FALSE;
extern struct db_watchpoint *db_watchpoint_list;
extern boolean_t db_watchpoints_inserted;


void
thread_kdb_return ()
{
  register thread_t self = current_thread ();
  register struct mc68020_saved_state *regs = self->pcb->user_regs;

  if (kdb_trap (regs->vector, regs))
    {
      thread_exception_return ();
      /*NOTREACHED*/
    }
}
#endif  MACH_KDB


static inline kern_return_t
kernel_page_fault (faultaddr, read_fault, regs)
        register vm_offset_t faultaddr;
        int read_fault;
        register struct frame *regs;
{
  register vm_map_t map;
  kern_return_t result;
  register thread_t thread = current_thread ();
  vm_prot_t fault_type;

  fault_type = read_fault ? VM_PROT_READ : VM_PROT_READ | VM_PROT_WRITE;

  /*
   * If the current map is a submap of the kernel map,
   * and the address is within that map, fault on that
   * map.  If we wait until vm_fault (vm_map_lookup) to
   * look for the submap, we may deadlock on the kernel
   * map lock.
   */
  if (thread == THREAD_NULL)
    map = kernel_map;
  else
    {
      map = thread->task->map;
      if (faultaddr < vm_map_min (map) || faultaddr >= vm_map_max (map))
        map = kernel_map;
    }

  result = vm_fault (map, trunc_page (faultaddr), fault_type, FALSE, FALSE,
                     (void (*) ()) 0);

#if     MACH_KDB
  if (result == KERN_SUCCESS && db_watchpoint_list
      && db_watchpoints_inserted && !read_fault)
    {
      if (db_find_watchpoint (map, faultaddr, regs))
        kdb_trap (T_WATCHPOINT, regs);
      else
        {
          if ((regs->sr & SR_T1) && (thread != THREAD_NULL))
            {
              spl7();
              thread->pcb->flag |= TRACE_KDB;
            }
        }
    }
#endif  MACH_KDB

  return result;
}


void
kernel_trap_c (int args)
{
  struct kern_frame *fr = (void*)&args;

  switch (fr->vector)
    {
    case V_BUSERR:
      if (current_thread () == THREAD_NULL)
        break;      /* trap in boot phase */

        if ((fr->u.fmtA.dfault)
            && (kernel_page_fault ((vm_offset_t)fr->u.fmtA.dfa,
                                   fr->u.fmtA.rw,fr) == KERN_SUCCESS))
              return;
        break;

      default:
          break;
    }
  asm ("illegal");
}


volatile void
thread_bootstrap_return ()
{
  volatile void thread_exception_return (void);

  asm ("illegal");
  thread_exception_return ();
}


volatile void
thread_exception_return (void)
{
  register pcb_t pcb = current_thread ()->pcb;

  asm ("illegal");

  if (!(pcb->user_regs->sr & 0x2000))
    while (need_ast[0])
      ast_taken ();

  if (pcb->discard_frame)
    {
     if(pcb->user_regs->format == FMTA || pcb->user_regs->format == FMTB)
       {
         register long *p asm ("a0") = (long*)pcb->user_regs;

         asm volatile (" movel   %0,sp;"
                       " clrw    sp@-;"
                       " movel   a0@+,a1;"
                       " movel   a1,usp;"
                       " movew   a0@(16*4),sp@-;"
                       " movel   a0@(15*4),sp@-;"
                       " moveml  a0@,d0-d7/a0-a6;"
                       " rte"
                       : : "r" (pcb->user_state));
       }
    }
  asm volatile (" movel   %0,sp;"
                " movel   sp@+,a0;"
                " movel   a0,usp;"
                " moveml  sp@+,d0-d7/a0-a6;"
                " rte"
                : : "r" (pcb->user_regs));
}


static volatile void
user_page_fault_continue (result)
     kern_return_t result;
{
  register thread_t thread = current_thread ();
  register pcb_t pcb = thread->pcb;
  register struct frame *regs = pcb->user_regs;

  if (result == KERN_SUCCESS) {
#if     MACH_KDB
    if (db_watchpoint_list && db_watchpoints_inserted
        && (pcb->fault_type & VM_PROT_WRITE))
      {
        if (db_find_watchpoint(thread->task->map, pcb->fault_addr, regs))
          kdb_trap (T_WATCHPOINT, regs);
      }
#endif  MACH_KDB
    /*
     *  If halt was requested, stop here instead of
     *  waiting for instruction to complete.  The page
     *  is still missing, and we`ll never finish.
     */
    while (thread_should_halt (thread))
      thread_halt_self ();

    thread_exception_return ();
    /*NOTREACHED*/
  }
  exception (EXC_BAD_ACCESS, result, pcb->fault_addr);
  /*NOTREACHED*/
}


static volatile inline void
user_page_fault (faultaddr, read_fault, regs)
     register vm_offset_t faultaddr;
     int read_fault;
     register struct frame *regs;
{
  register thread_t thread = current_thread ();
  register pcb_t pcb = thread->pcb;

  pcb->fault_addr = faultaddr;
  pcb->fault_type
    = read_fault ? VM_PROT_READ : VM_PROT_READ | VM_PROT_WRITE;
  vm_fault (thread->task->map, trunc_page (faultaddr), pcb->fault_type,
            FALSE, FALSE, user_page_fault_continue);
  /*NOTREACHED*/
}


/*
 *      User mode trap.
 */

volatile void
user_trap_c (struct frame *arg)
{
  register struct frame *regs = (&arg)[-1];
  register thread_t thread = current_thread ();
  register pcb_t pcb = thread->pcb;
  int exc, code, subcode = 0;

  pcb->user_regs = regs;
  pcb->discard_frame = FALSE;

  switch (code = regs->vector >> 2)
    {
    case T_BUSERR:
      /*
       * User bus error.  Try to handle FPA, then pagefault.
       */

      /*
       * Copy the "rerun" bits to the "fault" bits.
       *
       * This is what is going on here (don`t believe
       * the 2nd edition 68020 description in section
       * 6.4.1, it is full of errors).  A rerun bit
       * being on means that the prefetch failed.  A
       * fault bit being on means the processor tried
       * to use bad prefetch data.  Upon return via
       * the RTE instruction, the `20 will retry the
       * instruction access only if BOTH the rerun and
       * the corresponding fault bit is on.
       *
       * We need to do guarantee that any time we have a
       * fault that we have actually just run the cycle,
       * otherwise the current external state (i.e. the
       * bus error register) might not anything to do with
       * what really happened to cause the prefetch to fail.
       * For example the prefetch might have occured previous
       * to an earlier bus error exception whose handling
       * might have resolved the prefetch problem.  Thus by
       * copying the "rerun" bits, we force the `20 to rerun
       * every previously faulted prefetch upon return from
       * this bus error.  This way we are guaranteed that we
       * never get a "bogus" `20 internal bus error when it
       * attempts to use a previously faulted prefetch.  On
       * the downside, this hack might make the kernel fix up
       * a prefetch fault that the `20 was not going to use.
       * What we really need is a ``do not know anything about
       * a prefetch`` bit.  If we had something like that then
       * the `20 could know enough to rerun the prefetch, but
       * only if it turns out that it really needs it.
       *
       * RISC does have its advantages.
       *
       * N.B.  This code depends on not having an executable
       * where the last instruction in the text segment is
       * too close the end of a page.  We do not want to get
       * ourselves in trouble trying to fix up a fault beyond
       * the end of the text segment.  But because the loader
       * already pads out by an additional page when it sees
       * this problem due to microcode bugs with the first
       * year or so worth of 68020 chips, we should not get
       * in trouble here.
       */

      /* FIXME: ??? */
      regs->u.fmtA.faultc = regs->u.fmtA.rerunc;
      regs->u.fmtA.faultb = regs->u.fmtA.rerunb;

      if (regs->format == FMTA)
        {
          if (regs->u.fmtA.dfault)
            user_page_fault ((vm_offset_t)regs->u.fmtA.dfa,
                             regs->u.fmtA.rw, regs);
          else if (regs->u.fmtA.faultc)
            user_page_fault ((vm_offset_t)(regs->pc + 2), TRUE, regs);
          else if (regs->u.fmtA.faultb)
            user_page_fault ((vm_offset_t)(regs->pc + 4), TRUE, regs);
          /*NOTREACHED*/
        }
      else if (regs->format == FMTB)
        {
          if (regs->u.fmtB.dfault)
            user_page_fault ((vm_offset_t)regs->u.fmtB.dfa,
                             regs->u.fmtB.rw, regs);
          else if (regs->u.fmtB.faultc)
            user_page_fault ((vm_offset_t)(regs->u.fmtB.sba - 2), TRUE,
                             regs);
          else if (regs->u.fmtB.faultb)
            user_page_fault ((vm_offset_t)regs->u.fmtB.sba, TRUE, regs);
        }
      else
        panic ("bad bus error stack format");

      exc = EXC_BAD_ACCESS;
      break;

    case T_ZERODIV:
    case T_FLT_BSUN:
    case T_FLT_INEX:
    case T_FLT_ZERODIV:
    case T_FLT_UNFL:
    case T_FLT_OPERR:
    case T_FLT_OVFL:
    case T_FLT_SNAN:
      exc = EXC_ARITHMETIC;
      break;

    case T_CHKINST:             /* CHK [CHK2] instruction */
    case T_TRAPV:               /* TRAPV [cpTRAPcc TRAPcc] instr */
      exc = EXC_SOFTWARE;
      break;

    case T_EMU1010:             /* 1010 emulator trap */
    case T_EMU1111:             /* 1111 emulator trap */
      exc = EXC_EMULATION;
      break;

    case T_COPERR:           /* coprocessor protocol error */
      /*
       * Dump out obnoxious info to warn user
       * that something isn`t right w/ the 68881
       */
      /*
        showregs ("USER COPROCESSOR PROTOCOL ERROR", type, regs);
        */
    default:
      exc = EXC_BAD_INSTRUCTION;
      break;
    }
  exception (exc, code, subcode);
}


volatile void
syscall_c (struct frame *regs)
{
  thread_t thread = current_thread ();
  pcb_t pcb = thread->pcb;
  int exc, code = *((long*)&regs-1);
  int subcode = 0;
  mach_trap_t *trap;

  pcb->user_regs = regs;
  pcb->discard_frame = FALSE;

  if ((code > -10) || (code < -127)) /* Should be -129, but this allows a quicker moveq/compare */
    exception (EXC_SOFTWARE, code, 0);

  trap = &mach_trap_table[-code];
  if (trap->mach_trap_arg_count)
    {
      register int ret asm ("d0");

      asm (" moveml %0@,d2-d7/a5;"
           " moveml d2-d7/a5,sp@-;"
           " jbsr %1@;"
           " lea sp@(4*7),sp;"
           : /* no outputs (except ret in d0) */
           : "a" (regs->sp+4), "a" (trap->mach_trap_function)
           : "d2", "d3", "d4", "d5", "d6", "d7", "a5");
      thread_syscall_return (ret);
    }
  else
    thread_syscall_return ((*trap->mach_trap_function) ());
}


volatile void
thread_syscall_return (kern_return_t ret)
{
  struct frame *regs = current_thread ()->pcb->user_regs;
  register int _ret asm ("d0") = ret;

  asm ("illegal");

  if (!(regs->sr & 0x2000))
    while (need_ast[0])
      ast_taken ();

  asm volatile (" movel   %0@(-8),a0;"
                " movel   a0,usp;"
                " movel   %0,sp;"
                " moveml  sp@+,d1-d7/a0-a6;"
                " rte"
                : : "a" (&regs->d1) : "a0");
}


volatile void
mach_msg_reg_c (struct frame *arg)
{
  struct frame *regs = (&arg)[-1];
  thread_t thread = current_thread();
  pcb_t pcb = thread->pcb;

  pcb->user_regs = regs;
  pcb->discard_frame = FALSE;

  {
    register int ret asm ("d0");

    asm (" moveml %0@,d0-d5;"
         " moveml d0-d5,sp@-;"
         " movel %1,sp@-;"
         " jbsr _mach_msg_trap;"
         " lea sp@(4*7),sp;"
         : /* no outputs (except ret in d0) */
         : "a" (&regs->d2), "r" (regs->a2)
         : "d0", "d1", "d2", "d3", "d4", "d5");
    thread_syscall_return (ret);
  }
}


void
trap_init (void)
{
  /* Interface stubs in motorola/m68k/locore.s */
  void trap ();
  void syscall ();
  void interrupt ();
  void mach_msg_reg ();

  long *vector_table = get_vbr();

  vector_table[T_BUSERR] = (int)trap;
  vector_table[T_SYSCALL] = (int)syscall;
  vector_table[T_MACH_MSG_REG] = (int)mach_msg_reg;
  vector_table[T_INT0] = (int)interrupt;
  vector_table[T_INT1] = (int)interrupt;
  vector_table[T_INT2] = (int)interrupt;
  vector_table[T_INT3] = (int)interrupt;
  vector_table[T_INT4] = (int)interrupt;
  vector_table[T_INT5] = (int)interrupt;
  vector_table[T_INT6] = (int)interrupt;
  vector_table[T_INT7] = (int)interrupt;
}
